import argparse
import os
from typing import Set

from servey.util import get_servey_main


def generate_serverless_scaffold(skip_configs: Set[str]):
    main_serverless_yml_file = (
        os.environ.get("MAIN_SERVERLESS_YML_FILE") or "serverless.yml"
    )
    if not os.path.exists(main_serverless_yml_file):
        from servey.servey_aws.serverless import yml_config
        from importlib import resources

        template = resources.read_text(yml_config, "serverless_template.yml")
        serverless_yml_content = template.format(
            serverless_service_name=get_servey_main().title().replace("_", ""),
            service_name=get_servey_main(),
        )
        with open(main_serverless_yml_file, "w") as f:
            f.write(serverless_yml_content)
    from servey.servey_aws.serverless.yml_config.yml_config_abc import configure

    configure(main_serverless_yml_file, skip_configs)


def main():
    parser = argparse.ArgumentParser(description="Servey")
    parser.add_argument("--run", default="sls")
    parser.add_argument("--skip-configs", nargs="+", default=[])
    args = parser.parse_args()
    generate_serverless_scaffold(set(args.skip_configs))


if __name__ == "__main__":  # pragma: no cover
    main()



from dataclasses import dataclass
from http.client import HTTPException
from typing import Optional, Any

from marshy.types import ExternalItemType
from starlette.responses import Response, RedirectResponse

from servey.servey_starlette.action_endpoint.action_endpoint import ActionEndpoint
from servey.servey_web_page.redirect import Redirect
from servey.servey_web_page.web_page_trigger import get_environment


@dataclass
class WebPageActionEndpoint(ActionEndpoint):
    """
    Wrapper that combines an action with a template
    """

    template_name: Optional[str] = None
    content_type: str = "text/html"

    def __post_init__(self):
        if not self.template_name:
            self.template_name = f"{self.action.name}.j2"

    def render_response(self, result: Any):
        if isinstance(result, Redirect):
            return RedirectResponse(result.url, result.status_code)
        result_content = (
            self.result_marshaller.dump(result) if self.result_marshaller else None
        )
        if self.result_schema:
            error = next(self.result_schema.iter_errors(result_content), None)
            if error:
                raise HTTPException(500, str(error))
        body = self.template.render(model=result_content)
        return Response(content=body, media_type=self.content_type)

    @property
    def template(self):
        template = getattr(self, "_template", None)
        if not template:
            template = get_environment().get_template(self.template_name)
            setattr(self, "_template", template)
        return template

    def to_openapi_schema(self, schema: ExternalItemType):
        pass  # Not included in OpenAPI



from dataclasses import dataclass
from typing import Optional

from jinja2 import Environment, PackageLoader

from servey.trigger.trigger_abc import TriggerABC
from servey.trigger.web_trigger import WebTriggerMethod
from servey.util import get_servey_main


@dataclass(frozen=True)
class WebPageTrigger(TriggerABC):
    method: WebTriggerMethod = WebTriggerMethod.GET
    path: Optional[str] = None
    template_name: Optional[str] = None


def get_environment() -> Environment:
    environment = Environment(
        loader=PackageLoader(get_servey_main(), "templates"),
        auto_reload=False,  # Uvicorn handles this
    )
    environment.cache = {}
    return environment


PAGE_GET = WebPageTrigger()



from typing import Dict

from celery import Celery

from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.servey_celery.celery_config.celery_config_abc import CeleryConfigABC
from servey.trigger.fixed_rate_trigger import FixedRateTrigger


class FixedRateTriggerConfig(CeleryConfigABC):
    def configure(self, app: Celery, global_ns: Dict):
        has_triggers = False
        for _action, _ in find_actions_with_trigger_type(FixedRateTrigger):
            if _action.name not in global_ns:
                global_ns[_action.name] = app.task(_action.fn)
                has_triggers = True

        if not has_triggers:
            return

        # pylint: disable=W0613
        # noinspection PyUnusedLocal
        @app.on_after_configure.connect
        def setup_periodic_tasks(sender, **kwargs):
            for action, trigger in find_actions_with_trigger_type(FixedRateTrigger):
                task = global_ns[action.name]
                sender.add_periodic_task(trigger.interval, task.s())



import logging

from marshy.factory.impl_marshaller_factory import get_impls
from starlette.applications import Starlette

from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC

LOGGER = logging.getLogger(__name__)
routes = []
for route_factory in sorted(
    list(get_impls(RouteFactoryABC)), key=lambda f: f.priority, reverse=True
):
    routes.extend(route_factory().create_routes())
for route in routes:
    LOGGER.debug("starlette_path:%s", route.path)

app = Starlette(routes=routes)



from dataclasses import dataclass, field
from typing import Dict, Optional, Any

from marshy.types import ExternalItemType
from schemey import schema_from_type
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.routing import Route

from servey.action.action import Action
from servey.security.authorization import Authorization
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)
from servey.servey_starlette.error_response import ErrorResponse


@dataclass
class AuthorizingActionEndpoint(ActionEndpointABC):
    """
    Wrapper for a function for use within starlette, with everything needed to bind it to a route
    """

    action_endpoint: ActionEndpointABC
    authorizer: AuthorizerABC
    auth_kwarg_name: str
    error_schema: ExternalItemType = field(
        default_factory=lambda: schema_from_type(ErrorResponse).schema
    )

    def get_action(self) -> Action:
        return self.action_endpoint.get_action()

    def get_route(self) -> Route:
        route = self.action_endpoint.get_route()
        return Route(
            route.path, name=route.name, endpoint=self.execute, methods=route.methods
        )

    async def execute_with_context(
        self, request: Request, context: Dict[str, Any]
    ) -> Response:
        authorization = parse_authorization(self.authorizer, request)
        if not self.get_action().access_control.is_executable(authorization):
            return JSONResponse({"error": "unauthorized"}, 401)
        if self.auth_kwarg_name:
            context[self.auth_kwarg_name] = authorization
        response = await self.action_endpoint.execute_with_context(request, context)
        return response

    def to_openapi_schema(self, schema: ExternalItemType):
        self.action_endpoint.to_openapi_schema(schema)

        components: Dict = schema["components"]
        components["ErrorResponse"] = self.error_schema

        if not components.get("securitySchemas"):
            components["securitySchemes"] = {
                "OAuth2PasswordBearer": {
                    "type": "oauth2",
                    "flows": {"password": {"scopes": {}, "tokenUrl": "/login"}},
                }
            }

        route = self.action_endpoint.get_route()
        paths: Dict = schema["paths"]
        path: Dict = paths.get(route.path)
        if not path:
            return
        for method in route.methods:
            path_method: Dict = path.get(method.lower())
            if path_method is None:
                continue
            if not path_method.get("security"):
                path_method["security"] = [{"OAuth2PasswordBearer": []}]
            responses: Dict = path_method["responses"]
            responses["403"] = {
                "description": "unauthorized",
                "content": {
                    "application/json": {
                        "schema": {"$ref": "#components/ErrorResponse"}
                    }
                },
            }


def parse_authorization(
    authorizer: AuthorizerABC, request: Request
) -> Optional[Authorization]:
    token = request.headers.get("Authorization")
    if not token or not token.startswith("Bearer "):
        return
    token = token[7:]
    return authorizer.authorize(token)



import os
from dataclasses import dataclass
from pathlib import Path
from random import randint

from marshy.types import ExternalItemType
from ruamel.yaml import YAML

from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    create_yml_file,
)

ALPHABET = "0123456789abcdefghijklmnopqrstuvwxyz"


@dataclass
class StaticSiteBucketConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    static_site_directory: Path = Path(
        os.environ.get("SERVEY_STATIC_SITE_DIR") or "static_site"
    )
    static_site_bucket_resource_yml_file: str = (
        "serverless_servey/static_site_bucket_resource.yml"
    )

    def configure(self, main_serverless_yml_file: str):
        if not self.static_site_directory.exists():
            return
        # If there is a static site
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["resources"],
            self.static_site_bucket_resource_yml_file,
        )
        self.add_custom_bucket_yml(main_serverless_yml_file)
        static_site_bucket_resource_yml = self.build_static_site_bucket_resource_yml()
        create_yml_file(
            self.static_site_bucket_resource_yml_file, static_site_bucket_resource_yml
        )

    def add_custom_bucket_yml(self, main_serverless_yml_file: str):
        yaml = YAML()
        with open(main_serverless_yml_file, "r") as reader:
            root = yaml.load(reader)
            custom = root.get("custom")
            static_bucket = custom.get("staticBucket")
            if static_bucket:
                return static_bucket
        static_bucket = "b" + "".join(ALPHABET[randint(0, 35)] for _ in range(10))
        custom["staticBucket"] = static_bucket
        custom["s3Sync"] = [
            {
                "bucketName": "${self: custom.staticBucket}",
                "localDir": str(self.static_site_directory),
            }
        ]
        root["plugins"].append("serverless-s3-sync")
        with open(main_serverless_yml_file, "w") as writer:
            yaml.dump(root, writer)
        return static_bucket

    @staticmethod
    def build_static_site_bucket_resource_yml() -> ExternalItemType:
        static_site_bucket_resource = {
            "Resources": {
                "OriginAccessIdentity": {
                    "Type": "AWS::CloudFront::CloudFrontOriginAccessIdentity",
                    "Properties": {
                        "CloudFrontOriginAccessIdentityConfig": {
                            "Comment": "Allow cloudfront to access the static bucket"
                        }
                    },
                },
                "StaticBucket": {
                    "Type": "AWS::S3::Bucket",
                    "Properties": {
                        "AccessControl": "Private",
                        "BucketName": "${self:custom.staticBucket}",
                        "WebsiteConfiguration": {"IndexDocument": "index.html"},
                    },
                },
                "StaticSiteS3BucketPolicy": {
                    "Type": "AWS::S3::BucketPolicy",
                    "Properties": {
                        "Bucket": {"Ref": "StaticBucket"},
                        "PolicyDocument": {
                            "Statement": [
                                {
                                    "Sid": "AllowCloudFrontServicePrincipalReadOnly",
                                    "Effect": "Allow",
                                    "Principal": {
                                        "AWS": {
                                            "Fn::Join": [
                                                "",
                                                [
                                                    "arn:aws:iam::cloudfront:user/CloudFront Origin Access Identity ",
                                                    {"Ref": "OriginAccessIdentity"},
                                                ],
                                            ]
                                        }
                                    },
                                    "Action": ["s3:GetObject"],
                                    "Resource": {
                                        "Fn::Join": [
                                            "",
                                            [
                                                "arn:aws:s3:::",
                                                {"Ref": "StaticBucket"},
                                                "/*",
                                            ],
                                        ]
                                    },
                                }
                            ]
                        },
                    },
                },
            }
        }
        return static_site_bucket_resource



from dataclasses import dataclass
from datetime import datetime
from time import time

from marshy.types import ExternalItemType

from servey.cache_control.cache_control_abc import CacheControlABC
from servey.cache_control.cache_header import CacheHeader
from servey.cache_control.secure_hash_cache_control import SecureHashCacheControl


@dataclass(frozen=True)
class TtlCacheControl(CacheControlABC):
    ttl: int
    cache_control: CacheControlABC = SecureHashCacheControl()
    must_revalidate: bool = False

    def get_cache_header(self, item: ExternalItemType):
        cache_header = self.cache_control.get_cache_header(item)
        return self._wrap_cache_header(cache_header)

    def get_cache_header_from_content(self, content: bytes) -> CacheHeader:
        cache_header = self.cache_control.get_cache_header_from_content(content)
        return self._wrap_cache_header(cache_header)

    def _wrap_cache_header(self, cache_header: CacheHeader):
        expire_at = datetime.fromtimestamp(int(time()) + self.ttl)
        return CacheHeader(
            etag=cache_header.etag,
            updated_at=cache_header.updated_at,
            expire_at=expire_at,
            private=cache_header.private,
            must_revalidate=self.must_revalidate,
        )



import importlib
import logging
import pkgutil
from dataclasses import dataclass, field
from typing import Iterator

from servey.finder.subscription_finder_abc import SubscriptionFinderABC
from servey.subscription.subscription import Subscription
from servey.util import get_servey_main

LOGGER = logging.getLogger(__name__)


@dataclass
class ModuleSubscriptionFinder(SubscriptionFinderABC):
    """
    Default implementation of subscription finder which searches for channels in a particular module
    """

    root_module_name: str = field(
        default_factory=lambda: f"{get_servey_main()}.subscriptions"
    )

    def find_subscriptions(self) -> Iterator[Subscription]:
        try:
            module = importlib.import_module(self.root_module_name)
            # noinspection PyTypeChecker
            yield from _find_subscriptions_in_module(module)
        except ModuleNotFoundError as e:
            LOGGER.info(f"no_subscriptions_found:{e}")


def _find_subscriptions_in_module(module) -> Iterator[Subscription]:
    for value in module.__dict__.values():
        if isinstance(value, Subscription):
            yield value
    if not hasattr(module, "__path__"):
        return  # Module was not a package...
    paths = []
    paths.extend(module.__path__)
    module_infos = list(pkgutil.walk_packages(paths))
    for module_info in module_infos:
        sub_module_name = module.__name__ + "." + module_info.name
        sub_module = importlib.import_module(sub_module_name)
        # noinspection PyTypeChecker
        yield from _find_subscriptions_in_module(sub_module)



"""
Invoker for lambdas which assumes that action names are specified in environment variables, with one lambda per
action.
"""
import dataclasses
import importlib
import inspect
import json
import logging
import os
from marshy.types import ExternalItemType, ExternalType

from servey.action.action import get_action
from servey.errors import ServeyError
from servey.servey_aws.event_handler.event_handler_abc import get_event_handlers


def invoke(event: ExternalItemType, context) -> ExternalType:
    _LOGGER.info(json.dumps({"lambda_event": event}))
    for handler in _EVENT_HANDLERS:
        if handler.is_usable(event, context):
            return handler.handle(event, context)
    raise ServeyError("no_handler")


def find_action():
    """If we don't have a module and name, maybe we can use an ACTION_NAME"""
    action_module = importlib.import_module(os.environ["SERVEY_ACTION_MODULE"])
    action_function_name = os.environ["SERVEY_ACTION_FUNCTION_NAME"]
    if "." in action_function_name:
        action_class_name, action_function_name = action_function_name.split(".")
        action_class = getattr(action_module, action_class_name)
        action_function = getattr(action_class, action_function_name)
        action_ = get_action(action_function)
        fn = action_.fn

        def wrapper(*args, **kwargs):
            result = fn(*args, **kwargs)
            return result

        sig = inspect.signature(fn)
        sig = sig.replace(
            parameters=[
                p.replace(annotation=action_class) if p.name == "self" else p
                for p in sig.parameters.values()
            ]
        )
        wrapper.__signature__ = sig
        action_ = dataclasses.replace(action_, fn=wrapper)
        return action_

    action_function = getattr(action_module, action_function_name)
    action_ = get_action(action_function)
    return action_


_ACTION = find_action()
_EVENT_HANDLERS = get_event_handlers(_ACTION)
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)
logging.basicConfig(level=logging.INFO)



from dataclasses import dataclass
from typing import Type, Any, List

from marshy.types import ExternalItemType, ExternalType

from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import AuthorizationError
from servey.servey_aws.event_handler.event_handler import (
    EventHandler,
    EventHandlerFactory,
)
from servey.servey_aws.event_handler.event_handler_abc import EventHandlerABC
from servey.util import to_snake_case, attr_camel_case


class AppsyncEventHandler(EventHandler):
    def is_usable(self, event: ExternalType, context) -> bool:
        if isinstance(event, list):
            event = event[0]
        return "arguments" in event

    def parse_kwargs(self, event: ExternalItemType):
        arguments = {**event["arguments"]}
        source = event.get("source")
        if source is not None:
            arguments["self"] = source
        arguments = attrs_to_snake_case(arguments)
        if self.param_schema:
            self.param_schema.validate(arguments)
        kwargs = self.param_marshaller.load(arguments)
        if self.auth_kwarg_name or self.action.access_control != ALLOW_ALL:
            # noinspection PyTypeChecker
            headers: ExternalItemType = (event.get("request") or {}).get(
                "headers"
            ) or {}
            auth_header = headers.get("authorization")
            authorization = None
            if auth_header:
                if auth_header.lower().startswith("bearer "):
                    auth_header = auth_header[7:]
                authorization = self.authorizer.authorize(auth_header)
            if not self.action.access_control.is_executable(authorization):
                raise AuthorizationError("unauthorized")
            if self.auth_kwarg_name:
                kwargs[self.auth_kwarg_name] = authorization
        return kwargs

    def render_result(self, result: Any) -> ExternalType:
        result = super().render_result(result)
        result = attrs_to_camel_case(result)
        return result

    def handle_batch(self, events: List[ExternalItemType]) -> List[ExternalType]:
        results = super().handle_batch(events)
        results = [{"data": r} for r in results]
        return results


@dataclass
class AppsyncEventHandlerFactory(EventHandlerFactory):
    priority: int = 100
    event_handler_type: Type[EventHandlerABC] = AppsyncEventHandler


def attrs_to_snake_case(arguments: ExternalType) -> ExternalType:
    if isinstance(arguments, dict):
        arguments = {
            to_snake_case(k): attrs_to_snake_case(v) for k, v in arguments.items()
        }
    if isinstance(arguments, list):
        arguments = [attrs_to_snake_case(a) for a in arguments]
    return arguments


def attrs_to_camel_case(result: ExternalType) -> ExternalType:
    if isinstance(result, dict):
        result = {attr_camel_case(k): attrs_to_camel_case(v) for k, v in result.items()}
    if isinstance(result, list):
        result = [attrs_to_camel_case(a) for a in result]
    return result



from dataclasses import dataclass
from datetime import datetime
from email.utils import formatdate
from typing import Optional, Iterator

from schemey.util import filter_none

from servey.util import secure_hash


@dataclass(frozen=True)
class CacheHeader:
    etag: Optional[str] = None
    updated_at: Optional[datetime] = None
    expire_at: Optional[datetime] = None
    private: bool = True
    must_revalidate: bool = False

    def get_http_headers(self):
        return filter_none(
            {
                "ETag": self.etag,
                "Cache-Control": self.get_cache_control_str(),
                "Last-Modified": None
                if self.updated_at is None
                else formatdate(self.updated_at.timestamp(), usegmt=True),
                "Expires": None
                if self.expire_at is None
                else formatdate(self.expire_at.timestamp(), usegmt=True),
            }
        )

    def get_cache_control_str(self):
        directives = ["private" if self.private else "public"]
        if self.expire_at is not None:
            max_age = (self.expire_at - datetime.now()).seconds
            if max_age > 0:
                directives.append(f"max-age={max_age}")
        if self.must_revalidate:
            directives.append("must-revalidate")
        if len(directives) == 1:
            return "no-storage"
        return ",".join(directives)

    def combine_with(self, cache_headers: Iterator["CacheHeader"]) -> "CacheHeader":
        keys = [self.etag]
        updated_at = self.updated_at
        expire_at = self.expire_at
        private = True
        must_revalidate = False
        for sub_header in cache_headers:
            keys.append(sub_header.etag)
            if updated_at is not None:
                if sub_header.updated_at is None or sub_header.updated_at > updated_at:
                    updated_at = sub_header.updated_at
            if expire_at is not None:
                if sub_header.expire_at is None or sub_header.expire_at < expire_at:
                    expire_at = sub_header.expire_at
            private &= sub_header.private
            must_revalidate |= sub_header.must_revalidate
        cache_key = secure_hash(keys)
        return CacheHeader(cache_key, updated_at, expire_at, private, must_revalidate)



import dataclasses
import inspect
from typing import Optional

from marshy.factory.optional_marshaller_factory import get_optional_type
from marshy.types import ExternalType

from servey.action.action import Action, get_action
from servey.errors import ServeyError
from servey.servey_aws.event_handler.event_handler_abc import (
    get_event_handlers,
    EventHandlerABC,
)
from servey.servey_aws.router.router_abc import RouterABC
from servey.util import to_snake_case


class AppsyncRouter(RouterABC):
    priority: int = 110

    def create_handler(self, event: ExternalType, context) -> EventHandlerABC:
        if isinstance(event, list):
            event = event[0]
        info = event.get("info", None)  # Diff appsync events
        if info is None:
            return
        field_name = info["fieldName"]
        source = event.get("source")
        if source:
            action = self.find_action_for_parent_type(
                info["parentTypeName"], field_name
            )
        else:
            action = self.find_action_for_field_name(field_name)
        if action is None:
            raise ServeyError(f"unknown_field_name:{field_name}")
        handlers = get_event_handlers(action)
        for handler in handlers:
            if handler.is_usable(event, context):
                return handler

    def find_action_for_parent_type(self, parent_type_name: str, field_name: str):
        field_name = to_snake_case(field_name)
        for action, _ in self.web_trigger_actions:
            sig = inspect.signature(action.fn)
            parent_type = sig.return_annotation
            parent_type = get_optional_type(parent_type) or parent_type
            if getattr(parent_type, "__name__", None) == parent_type_name:
                fn = getattr(parent_type, field_name)
                nested_action = get_action(fn)
                sig = inspect.signature(fn)
                parameters = list(sig.parameters.values())
                parameters[0] = parameters[0].replace(annotation=parent_type)
                sig = sig.replace(parameters=parameters)

                # pylint: disable=W0640
                def wrapper(*args, **kwargs):
                    return fn(*args, **kwargs)

                wrapper.__signature__ = sig
                nested_action = dataclasses.replace(nested_action, fn=wrapper)
                wrapper.__servey_action__ = nested_action
                return nested_action

    def find_action_for_field_name(self, field_name: str) -> Optional[Action]:
        for action, _ in self.web_trigger_actions:
            action_field_name = action.name[0] + action.name.title()[1:].replace(
                "_", ""
            )
            if action_field_name == field_name:
                return action



from typing import Optional

from marshy.types import ExternalItemType

from servey.action.action import Action
from servey.errors import ServeyError
from servey.servey_aws.event_handler.event_handler_abc import (
    get_event_handlers,
    EventHandlerABC,
)
from servey.servey_aws.router.router_abc import RouterABC


class APIGatewayRouter(RouterABC):
    priority: int = 120

    def create_handler(self, event: ExternalItemType, context) -> EventHandlerABC:
        if not isinstance(event, dict):
            return
        path = event.get("path", None)
        if path is None:
            return
        action = self.find_action_for_path(path)
        if action is None:
            raise ServeyError(f"unknown_path:{path}")
        handlers = get_event_handlers(action)
        for handler in handlers:
            if handler.is_usable(event, context):
                return handler

    def find_action_for_path(self, path: str) -> Optional[Action]:
        for action, trigger in self.web_trigger_actions:
            action_path = trigger.path or f"/actions/{action.name.replace('_', '-')}"
            if action_path == path:
                return action



import dataclasses
import inspect
from typing import Tuple

from servey.action.action import Action
from servey.servey_strawberry.handler_filter.handler_filter_abc import (
    HandlerFilterABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory


class StrawberryTypeHandlerFilter(HandlerFilterABC):
    """
    Sometimes the default behaviour from strawberry is not quite what we want, so this allows us to customize
    how a strawberry object is converted to / from regular python
    Examples include adding authorization
    """

    priority: int = 50

    def filter(
        self,
        action: Action,
        schema_factory: SchemaFactory,
    ) -> Tuple[Action, bool]:
        fn = action.fn
        sig = inspect.signature(fn)
        params = [
            p.replace(annotation=schema_factory.get_input(p.annotation))
            for p in sig.parameters.values()
        ]
        sig = sig.replace(
            parameters=params,
            return_annotation=schema_factory.get_type(sig.return_annotation),
        )

        def resolver(*args, **kwargs):
            # Should we unwrap strawberry inputs here?
            result = fn(*args, **kwargs)
            # Should we wrap the result type here?
            return result

        resolver.__signature__ = sig
        wrapped_action = dataclasses.replace(action, fn=resolver)

        return wrapped_action, True



import asyncio
import inspect
import mimetypes
from dataclasses import field, dataclass
from typing import Optional, Awaitable

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType
from schemey import get_default_schema_context, SchemaContext

from servey.action.action import Action
from servey.action.util import get_marshaller_for_params, get_schema_for_params
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_aws.event_handler.api_gateway_event_handler import (
    ApiGatewayEventHandler,
)
from servey.servey_aws.event_handler.event_handler import separate_auth_kwarg
from servey.servey_aws.event_handler.event_handler_abc import EventHandlerFactoryABC
from servey.servey_web_page.redirect import Redirect
from servey.servey_web_page.web_page_trigger import WebPageTrigger, get_environment


@dataclass
class WebPageEventHandler(ApiGatewayEventHandler):
    template_name: Optional[str] = None
    content_type: str = "text/html"

    @property
    def template(self):
        template = getattr(self, "_template", None)
        if not template:
            template = get_environment().get_template(self.template_name)
            setattr(self, "_template", template)
        return template

    def handle(self, event: ExternalItemType, context) -> ExternalItemType:
        kwargs = self.parse_kwargs(event)
        result = self.action.fn(**kwargs)
        if isinstance(result, Awaitable):
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(result)
        if isinstance(result, Redirect):
            return {
                "statusCode": result.status_code,
                "headers": {"Location": result.url},
            }

        dumped = self.result_marshaller.dump(result)
        body = self.template.render(model=dumped)
        headers = {}
        if self.content_type:
            headers["Content-Type"] = self.content_type
        response = {
            "statusCode": 200,
            "headers": headers,
            "body": body,
        }
        self.apply_caching(event, response)
        return response


@dataclass
class WebPageEventHandlerFactory(EventHandlerFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    auth_kwarg_name: Optional[str] = None
    authorizer: Optional[AuthorizerABC] = None
    validate_output: bool = True
    priority: int = 110

    def create(self, action: Action) -> Optional[WebPageEventHandler]:
        trigger = next(
            (t for t in action.triggers if isinstance(t, WebPageTrigger)), None
        )
        if not trigger:
            return

        fn, auth_kwarg_name = separate_auth_kwarg(action.fn)
        param_marshaller = get_marshaller_for_params(fn, set(), self.marshaller_context)
        result_marshaller = None
        param_schema = get_schema_for_params(fn, set(), self.schema_context)
        result_schema = None
        sig = inspect.signature(fn)
        if sig.return_annotation != inspect.Signature.empty:
            result_marshaller = self.marshaller_context.get_marshaller(
                sig.return_annotation
            )
            if self.validate_output:
                result_schema = self.schema_context.schema_from_type(
                    sig.return_annotation
                )
        authorizer = (
            get_default_authorizer()
            if auth_kwarg_name or action.access_control != ALLOW_ALL
            else None
        )
        return WebPageEventHandler(
            action=action,
            param_marshaller=param_marshaller,
            param_schema=param_schema,
            result_marshaller=result_marshaller,
            result_schema=result_schema,
            auth_kwarg_name=auth_kwarg_name,
            authorizer=authorizer,
            priority=self.priority,
            template_name=trigger.template_name or f"{action.name}.j2",
            content_type=mimetypes.guess_type(action.name or "")[0] or "text/html",
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import List, Optional, Set

from servey.action.action import Action
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)


class ActionEndpointFactoryABC(ABC):
    priority: 100

    @abstractmethod
    def create(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        """Create an action endpoint"""


def create_endpoint_factories() -> List[ActionEndpointFactoryABC]:
    from marshy.factory.impl_marshaller_factory import get_impls

    factories = [f() for f in get_impls(ActionEndpointFactoryABC)]
    factories.sort(key=lambda f: f.priority, reverse=True)
    return factories



from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Optional

from servey.security.authorization import Authorization

T = TypeVar("T")


class EventFilterABC(ABC, Generic[T]):
    @abstractmethod
    def should_publish(self, event: T, authorization: Optional[Authorization]) -> bool:
        """Filter the event given depending on the authorization of the subscriber"""



import json
from dataclasses import dataclass, field
from datetime import datetime

from marshy import get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

from servey.cache_control.cache_control_abc import CacheControlABC
from servey.cache_control.cache_header import CacheHeader
from servey.cache_control.secure_hash_cache_control import SecureHashCacheControl


@dataclass(frozen=True)
class TimestampCacheControl(CacheControlABC):
    cache_control: CacheControlABC = SecureHashCacheControl()
    updated_at_attr: str = "updated_at"
    timestamp_marshaller: MarshallerABC[datetime] = field(
        default_factory=lambda: get_default_context().get_marshaller(datetime)
    )

    def get_cache_header(self, item: ExternalItemType):
        cache_header = self.cache_control.get_cache_header(item)
        updated_at = item.get(self.updated_at_attr)
        updated_at = self.timestamp_marshaller.load(updated_at)
        return CacheHeader(
            etag=cache_header.etag,
            updated_at=updated_at,
            expire_at=cache_header.expire_at,
            private=cache_header.private,
        )

    def get_cache_header_from_content(self, content: bytes) -> CacheHeader:
        content_json = json.loads(content)
        return self.get_cache_header(content_json)



from marshy.types import ExternalItemType

from servey.action.action import Action
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)
from servey.servey_web_page.web_page_trigger import WebPageTrigger
from servey.trigger.trigger_abc import TriggerABC


class WebPageTriggerHandler(TriggerHandlerABC):
    def handle_trigger(
        self, action: Action, trigger: TriggerABC, lambda_definition: ExternalItemType
    ):
        if not isinstance(trigger, WebPageTrigger):
            return
        events = lambda_definition.get("events")
        if not events:
            events = lambda_definition["events"] = []
        path = trigger.path or f"/{action.name.replace('_', '-')}"
        events.append(
            {"http": {"path": path, "method": trigger.method.value, "cors": True}}
        )



import base64
import hashlib
import json
import os
import re

from marshy.types import ExternalType

_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")


def to_snake_case(name: str) -> str:
    return _PATTERN.sub("_", name).lower()


def entity_to_camel_case(name: str) -> str:
    parts = name.split("_")
    result = []
    for part in parts:
        result.append(part[0].upper())
        result.append(part[1:].lower())
    return "".join(result)


def attr_camel_case(name: str) -> str:
    parts = name.split("_")
    result = [parts[0].lower()]
    for part in parts[1:]:
        result.append(part[0].upper())
        result.append(part[1:].lower())
    return "".join(result)


def secure_hash(item: ExternalType) -> str:
    item_json = json.dumps(item)
    item_bytes = item_json.encode("utf-8")
    return secure_hash_content(item_bytes)


def secure_hash_content(content: bytes) -> str:
    sha = hashlib.sha256()
    sha.update(content)
    hash_bytes = sha.digest()
    b64_bytes = base64.b64encode(hash_bytes)
    b64_str = b64_bytes.decode("utf-8")
    return b64_str


def get_servey_main():
    servey_main = os.environ.get("SERVEY_MAIN")
    if servey_main:
        servey_main = to_snake_case(servey_main)
    else:
        servey_main = "servey_main"
    return servey_main



import dataclasses
import inspect
import os
from dataclasses import dataclass, field
from datetime import datetime
from typing import Type, Set

from marshy.factory.optional_marshaller_factory import get_optional_type
from marshy.types import ExternalItemType
from ruamel.yaml import YAML

from servey.action.action import Action, get_action
from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.trigger.web_trigger import WebTrigger, WebTriggerMethod
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    GENERATED_HEADER,
    create_yml_file,
)
from servey.util import get_servey_main, attr_camel_case


@dataclass
class AppsyncConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    servey_actions_yml_file: str = "serverless_servey/appsync.yml"
    servey_appsync_schema_file: str = "serverless_servey/schema.graphql"
    use_router_for_all: bool = field(
        default_factory=lambda: int(os.environ.get("SERVEY_AWS_ROUTER_FOR_ALL", "0"))
        == 1
    )

    def configure(self, main_serverless_yml_file: str):
        yaml = YAML()
        with open(main_serverless_yml_file, "r") as reader:
            root = yaml.load(reader)
            root["appSync"] = "${file(" + self.servey_actions_yml_file + ")}"
        with open(main_serverless_yml_file, "w") as writer:
            yaml.dump(root, writer)
        self.build_appsync_schema()
        servey_action_functions_yml = self.build_servey_action_functions_yml()
        create_yml_file(self.servey_actions_yml_file, servey_action_functions_yml)

    def build_appsync_schema(self):
        from servey.servey_strawberry.schema_factory import create_schema

        with open(self.servey_appsync_schema_file, "w") as writer:
            writer.write("# ")
            writer.write(GENERATED_HEADER.replace("\n", "\n# "))
            writer.write(f"\n# Updated at: {datetime.now().isoformat()}\n\n")
            schema = create_schema()

            schema = str(schema)

            # Using string manipulation here feels really gross, but I have not been able to
            # get directives working in strawberry...

            # noinspection SpellCheckingInspection
            schema = schema.replace(
                '"""Date with time (isoformat)"""\nscalar DateTime\n', ""
            )
            schema = schema.replace("scalar UUID\n", "")
            # Ugly AWS variable substitution
            schema = schema.replace(": DateTime", ": AWSDateTime")
            schema = schema.replace(": UUID", ": ID")
            writer.write(schema)

    def build_servey_action_functions_yml(self) -> ExternalItemType:
        appsync_definitions = {
            "name": get_servey_main(),
            "authentication": {
                "type": "API_KEY",
            },
            "apiKeys": [
                {
                    "name": f"{get_servey_main()}Key",
                }
            ],
            "resolvers": {},
            "dataSources": {},
            "schema": self.servey_appsync_schema_file,
        }
        processed_dataclasses = set()
        for action, trigger in find_actions_with_trigger_type(WebTrigger):
            resolver_type = (
                "Query" if trigger.method == WebTriggerMethod.GET else "Mutation"
            )
            resolver_name = resolver_type + "." + attr_camel_case(action.name)
            self.add_field(action, appsync_definitions, resolver_name)
            self.build_nested_resolvers(
                action, appsync_definitions, processed_dataclasses
            )

        return appsync_definitions

    def add_field(
        self, action: Action, appsync_definitions: ExternalItemType, resolver_name: str
    ):
        # data source may be
        use_router = self.use_router_for_all or "<locals>" in action.fn.__qualname__
        data_source_name = "servey_router" if use_router else action.name
        resolver = {
            "kind": "UNIT",
            "dataSource": data_source_name,
        }
        if action.batch_invoker:
            resolver["maxBatchSize"] = action.batch_invoker.max_batch_size
        appsync_definitions["resolvers"][resolver_name] = resolver
        data_source = {
            "type": "AWS_LAMBDA",
            "config": {
                "functionName": data_source_name,
            },
        }
        if action.description and not use_router:
            data_source["description"] = action.description.strip()
        appsync_definitions["dataSources"][data_source_name] = data_source

    def build_nested_resolvers(
        self,
        action: Action,
        appsync_definitions: ExternalItemType,
        processed_dataclasses: Set[Type],
    ):
        sig = inspect.signature(action.fn)
        type_ = sig.return_annotation
        type_ = get_optional_type(type_) or type_
        if not dataclasses.is_dataclass(type_) or type_ in processed_dataclasses:
            return
        processed_dataclasses.add(type_)
        members = inspect.getmembers(type_)
        for name, member in members:
            nested_action = get_action(member)
            if nested_action:
                resolver_name = type_.__name__ + "." + attr_camel_case(name)
                self.add_field(nested_action, appsync_definitions, resolver_name)
                self.build_nested_resolvers(
                    nested_action, appsync_definitions, processed_dataclasses
                )



from typing import Optional, List, Type

import typing_inspect

from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory

_TYPES_BY_ORIGIN = {
    # t.__origin__: t for t in typing.__dict__.values() if hasattr(t, "__origin__")
    list: List,
    set: List,
    frozenset: List,
}


class GenericFactory(EntityFactoryABC):
    priority: int = 160

    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        origin = typing_inspect.get_origin(annotation)
        if origin:
            origin = _TYPES_BY_ORIGIN.get(origin) or origin
            args = tuple(
                schema_factory.get_type(a) for a in typing_inspect.get_args(annotation)
            )
            # noinspection PyTypeChecker
            return origin[args]

    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        origin = typing_inspect.get_origin(annotation)
        if origin:
            origin = _TYPES_BY_ORIGIN.get(origin) or origin
            args = tuple(
                schema_factory.get_input(a) for a in typing_inspect.get_args(annotation)
            )
            # noinspection PyTypeChecker
            return origin[args]



"""
App for case where servey is run in local schedule mode. Scheduled events are run using
threads (As opposed to Celery).
Run this with `python -m servey.servey_thread`
"""
import logging
import os

from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.trigger.fixed_rate_trigger import FixedRateTrigger
from servey.servey_thread.fixed_rate_trigger_thread import FixedRateTriggerThread

LOGLEVEL = os.environ.get("LOGLEVEL", "INFO").upper()
logging.basicConfig(level=LOGLEVEL)
_LOGGER = logging.getLogger(__name__)
DAEMON = (os.environ.get("SERVEY_DAEMON") or "").lower() in ("t", "true", "1")
_THREADS = []
_LOGGER.info("Starting threads for actions...")
for action, trigger in find_actions_with_trigger_type(FixedRateTrigger):
    _LOGGER.info(f"Starting: {action.name}")
    t = FixedRateTriggerThread(action.fn, trigger, DAEMON)
    _THREADS.append(t)
    t.start()



import inspect
import mimetypes
from dataclasses import field, dataclass
from typing import Set, List, Optional

from marshy import get_default_context
from marshy.marshaller_context import MarshallerContext
from schemey import SchemaContext, get_default_schema_context

from servey.action.action import Action
from servey.action.util import get_marshaller_for_params, get_schema_for_params
from servey.errors import ServeyError
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
    ActionEndpointFactoryABC,
)
from servey.servey_web_page.web_page_action_endpoint import WebPageActionEndpoint
from servey.servey_web_page.web_page_trigger import WebPageTrigger


@dataclass
class WebPageActionEndpointFactory(ActionEndpointFactoryABC):
    """
    Factory for endpoints using templates
    """

    priority: int = 100
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    validate_output: bool = True
    path_pattern: str = "/actions/{action_name}"

    def create(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[WebPageActionEndpoint]:
        triggers = [t for t in action.triggers if isinstance(t, WebPageTrigger)]
        if not triggers:
            return
        if len(triggers) > 1:
            raise ServeyError(f"multi_triggers_not_supported:{triggers}")
        trigger = triggers[0]
        path = trigger.path or f"/{action.name.replace('_', '-')}"
        result_type = inspect.signature(action.fn).return_annotation
        if result_type == inspect.Signature.empty:
            result_type = None
        endpoint = WebPageActionEndpoint(
            action=action,
            path=path,
            methods=(trigger.method,),
            params_marshaller=get_marshaller_for_params(
                action.fn, skip_args, self.marshaller_context
            ),
            params_schema=get_schema_for_params(
                action.fn, skip_args, self.schema_context
            ),
            result_marshaller=self.marshaller_context.get_marshaller(result_type)
            if result_type
            else None,
            result_schema=self.schema_context.schema_from_type(result_type)
            if self.validate_output and result_type
            else None,
            template_name=trigger.template_name,
            content_type=mimetypes.guess_type(action.name or "")[0] or "text/html",
        )
        return endpoint



"""
Invoker which groups lambdas together in such a way that one lambda may be used to invoke multiple
different actions. Routers are used to try and link an event to an action, which may be specific
to APIGateway, AppSync, or Direct Invocation
"""
import json
import logging
from marshy.types import ExternalItemType, ExternalType

from servey.errors import ServeyError
from servey.servey_aws.router.router_abc import find_routers


def invoke(event: ExternalItemType, context) -> ExternalType:
    _LOGGER.info(json.dumps({"lambda_event": event}))
    for router in _ROUTERS:
        handler = router.create_handler(event, context)
        if handler:
            return handler.handle(event, context)
    raise ServeyError("no_handler")


_ROUTERS = find_routers()
_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)
logging.basicConfig(level=logging.INFO)



from abc import ABC, abstractmethod
from typing import Optional, Tuple, List

from marshy.factory.impl_marshaller_factory import get_impls

from servey.finder.subscription_finder_abc import find_subscriptions

# pylint: disable=R0401
from servey.subscription.subscription import Subscription, T


class SubscriptionServiceABC(ABC):
    """
    Service for sending messages to subscribed clients (Possibly by websocket)
    """

    @abstractmethod
    def publish(self, subscription: Subscription[T], event: T):
        """
        Send the event given to listeners on the subscription given.
        """


class SubscriptionServiceFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create(
        self, subscriptions: List[Subscription]
    ) -> Optional[SubscriptionServiceABC]:
        """
        Create a subscription service if the current environment supports it.
        """


_SUBSCRIPTION_SERVICES = None


# pylint: disable=W0603
def get_subscription_services() -> Tuple[SubscriptionServiceABC, ...]:
    global _SUBSCRIPTION_SERVICES
    if _SUBSCRIPTION_SERVICES is None:
        subscriptions = list(find_subscriptions())
        _SUBSCRIPTION_SERVICES = tuple(
            s
            for s in (
                f().create(subscriptions)
                for f in get_impls(SubscriptionServiceFactoryABC)
            )
            if s
        )
    return _SUBSCRIPTION_SERVICES



import base64
import dataclasses
import logging
import os
from dataclasses import field
from typing import Optional

from servey.security.authenticator.password_authenticator_abc import (
    PasswordAuthenticatorABC,
)
from servey.security.authorization import ROOT, Authorization

LOGGER = logging.getLogger(__name__)


def _default_password():
    password = os.environ.get("SERVEY_DEBUG_AUTHENTICATOR_PASSWORD")
    if not password:
        password = (
            base64.b64encode(os.urandom(12))
            .decode("UTF-8")
            .replace("+", "")
            .replace("/", "")
        )
        LOGGER.warning(f"Using Temporary Password: {password}")
    return password


@dataclasses.dataclass
class RootPasswordAuthenticator(PasswordAuthenticatorABC):
    username: str = field(
        default_factory=lambda: os.environ.get("DEBUG_AUTHENTICATOR_USERNAME") or "root"
    )
    password: str = field(default_factory=_default_password)
    authorization: Authorization = field(default=ROOT)

    def authenticate(self, username: str, password: str) -> Optional[Authorization]:
        if self.username == username and self.password == password:
            return self.authorization



from dataclasses import field, dataclass
from typing import Optional

from marshy import ExternalType
from marshy.types import ExternalItemType


@dataclass
class Example:
    """Example invocation which may be used for mocks and unit testing."""

    name: str
    params: ExternalItemType = field(default_factory=dict)
    result: ExternalType = None
    description: Optional[str] = None
    include_in_tests: bool = True
    include_in_schema: bool = True



from typing import Dict

from celery import Celery

from servey.finder.subscription_finder_abc import find_subscriptions
from servey.servey_celery.celery_config.celery_config_abc import CeleryConfigABC


class SubscriptionConfig(CeleryConfigABC):
    def configure(self, app: Celery, global_ns: Dict):
        for _subscription in find_subscriptions():
            if _subscription.action_subscribers:
                for _action in _subscription.action_subscribers:
                    if _action.name not in global_ns:
                        global_ns[_action.name] = app.task(_action.fn)



from marshy.types import ExternalType

from servey.finder.action_finder_abc import find_actions
from servey.servey_aws.event_handler.event_handler_abc import (
    get_event_handlers,
    EventHandlerABC,
)
from servey.servey_aws.router.router_abc import RouterABC


class Router(RouterABC):
    def create_handler(self, event: ExternalType, context) -> EventHandlerABC:
        if isinstance(event, list):
            event = event[0]
        action_name = event.get("action_name", None)
        if action_name is None:
            return
        action = next(a for a in find_actions() if a.name == action_name)
        handlers = get_event_handlers(action)
        for handler in handlers:
            if handler.is_usable(event, context):
                return handler



from marshy.types import ExternalItemType

from servey.action.action import Action
from servey.trigger.trigger_abc import TriggerABC
from servey.trigger.web_trigger import WebTrigger
from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
    TriggerHandlerABC,
)


class WebTriggerHandler(TriggerHandlerABC):
    """
    Handler which outputs serverless definitions for api gateway triggers.
    """

    path_pattern: str = "/actions/{action_name}"

    def handle_trigger(
        self,
        action: Action,
        trigger: TriggerABC,
        lambda_definition: ExternalItemType,
    ):
        if not isinstance(trigger, WebTrigger):
            return
        events = lambda_definition.get("events")
        if not events:
            events = lambda_definition["events"] = []
        path = trigger.path or self.path_pattern.format(
            action_name=action.name.replace("_", "-")
        )
        events.append(
            {"http": {"path": path, "method": trigger.method.value, "cors": True}}
        )



import os
from dataclasses import field, dataclass
from typing import Iterator

from marshy.types import ExternalItemType
from schemey.util import filter_none
from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.routing import Route, Mount
from starlette.staticfiles import StaticFiles

from servey.finder.action_finder_abc import find_actions
from servey.servey_starlette.route_factory.action_route_factory import (
    ActionRouteFactory,
)
from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC


@dataclass
class OpenapiRouteFactory(RouteFactoryABC):
    title: str = field(
        default_factory=lambda: os.environ.get("SERVEY_API_TITLE") or "Servey"
    )
    description: str = field(
        default_factory=lambda: os.environ.get("SERVEY_API_DESCRIPTION")
    )
    version: str = field(
        default_factory=lambda: os.environ.get("SERVEY_API_VERSION") or "0.1.0"
    )
    action_route_factory: ActionRouteFactory = field(default_factory=ActionRouteFactory)
    debug: bool = field(
        default_factory=lambda: int(os.environ.get("SERVER_DEBUG", "1")) == 1
    )

    def create_routes(self) -> Iterator[Route]:
        if not self.debug:
            return
        # add as template route
        yield Route("/openapi.json", endpoint=self.endpoint, include_in_schema=False)
        yield Mount(
            "/docs",
            app=StaticFiles(packages=["servey.servey_starlette"], html=True),
            name="docs",
        )

    # pylint: disable=W0613
    # noinspection PyUnusedLocal
    def endpoint(self, request: Request) -> Response:
        schema = self.openapi_schema()
        return JSONResponse(schema)

    def openapi_schema(self) -> ExternalItemType:
        schema = {
            "openapi": "3.0.2",
            "info": filter_none(
                {
                    "title": self.title,
                    "version": self.version,
                    "description": self.description,
                }
            ),
            "paths": {},
            "components": {},
        }
        if os.environ.get("SERVER_HOST"):
            schema["servers"] = [{"url": os.environ.get("SERVER_HOST")}]
        for action in find_actions():
            action_endpoint = self.action_route_factory.create_action_endpoint(action)
            if action_endpoint:
                action_endpoint.to_openapi_schema(schema)
        return schema



import dataclasses
import importlib
import inspect
from dataclasses import dataclass
from typing import List, Optional, Set

from servey.action.action import Action
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
    ActionEndpointFactoryABC,
)


@dataclass
class SelfActionEndpointFactory(ActionEndpointFactoryABC):
    """
    Handles case where `self` is unannotated - wraps the action function adding an annotation
    """

    priority: int = 250
    skip: bool = False

    def create(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        if self.skip:
            return
        sig = inspect.signature(action.fn)
        params = list(sig.parameters.values())
        if (
            not params
            or params[0].name != "self"
            or params[0].annotation != inspect.Signature.empty
        ):
            return

        # noinspection PyUnresolvedReferences
        subject_module = importlib.import_module(action.fn.__module__)
        subject_class_name = action.fn.__qualname__.split(".")[0]
        subject_class = getattr(subject_module, subject_class_name)

        def wrapper(*args, **kwargs):
            result = action.fn(*args, **kwargs)
            return result

        params[0] = params[0].replace(annotation=subject_class)
        wrapper.__signature__ = sig.replace(parameters=params)
        wrapped_action = dataclasses.replace(action, fn=wrapper)
        action_endpoint = self._get_wrapped_endpoint(
            wrapped_action, skip_args, factories
        )
        return action_endpoint

    def _get_wrapped_endpoint(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        self.skip = True
        try:
            for factory in factories:
                action_endpoint = factory.create(action, skip_args, factories)
                if action_endpoint:
                    return action_endpoint
        finally:
            self.skip = False



import asyncio
import inspect
from dataclasses import field, dataclass
from typing import Optional, Tuple, Callable, Any, Type, Awaitable, List

from marshy import get_default_context, ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshaller_context import MarshallerContext
from marshy.types import ExternalItemType
from marshy.utils import resolve_forward_refs
from schemey import get_default_schema_context, SchemaContext, Schema

from servey.action.action import Action
from servey.action.util import get_marshaller_for_params, get_schema_for_params
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import (
    Authorization,
    get_inject_at,
    AuthorizationError,
)
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_aws.event_handler.event_handler_abc import (
    EventHandlerABC,
    EventHandlerFactoryABC,
)


# pylint: disable=R0902
@dataclass
class EventHandler(EventHandlerABC):
    action: Action
    param_marshaller: MarshallerABC
    param_schema: Optional[Schema]
    result_marshaller: Optional[MarshallerABC]
    result_schema: Optional[Schema] = None
    auth_kwarg_name: Optional[str] = None
    auth_marshaller: Optional[MarshallerABC[Authorization]] = None
    authorizer: Optional[AuthorizerABC] = None
    priority: int = 50

    def is_usable(self, event: ExternalType, context) -> bool:
        if isinstance(event, list):
            if not self.action.batch_invoker:
                return False
            event = event[0]
        return "params" in event

    def handle(self, event: ExternalType, context) -> ExternalType:
        if isinstance(event, list):
            return self.handle_batch(event)
        return self.handle_event(event)

    def handle_batch(self, events: List[ExternalItemType]) -> List[ExternalType]:
        if not events:
            return []
        kwarg_list = [self.parse_kwargs(e) for e in events]
        arg_extractor = self.action.batch_invoker.arg_extractor
        key_list = [arg_extractor(a["self"])[0] for a in kwarg_list]
        authorization = kwarg_list[0]["authorization"]
        results = self.action.batch_invoker.fn(key_list, authorization)
        if isinstance(results, Awaitable):
            loop = asyncio.get_event_loop()
            results = loop.run_until_complete(results)
        results = [self.render_result(result) for result in results]
        return results

    def handle_event(self, event: ExternalItemType) -> ExternalType:
        kwargs = self.parse_kwargs(event)
        result = self.action.fn(**kwargs)
        if isinstance(result, Awaitable):
            loop = asyncio.get_event_loop()
            result = loop.run_until_complete(result)
        return self.render_result(result)

    def parse_kwargs(self, event: ExternalItemType):
        params = event["params"]
        if self.param_schema:
            self.param_schema.validate(params)
        kwargs = self.param_marshaller.load(params)
        if self.action.access_control != ALLOW_ALL or self.auth_kwarg_name:
            auth_kwarg_value = event.get("authorization")
            authorization = None
            if isinstance(auth_kwarg_value, str):
                authorization = self.authorizer.authorize(auth_kwarg_value)
            elif isinstance(auth_kwarg_value, dict):
                authorization = self.auth_marshaller.load(auth_kwarg_value)
            if not self.action.access_control.is_executable(authorization):
                raise AuthorizationError("unauthorized")
            if self.auth_kwarg_name:
                kwargs[self.auth_kwarg_name] = authorization
        return kwargs

    def render_result(self, result: Any) -> ExternalType:
        if self.result_marshaller:
            result = self.result_marshaller.dump(result)
        if self.result_schema:
            self.result_schema.validate(result)
        return result


@dataclass
class EventHandlerFactory(EventHandlerFactoryABC):
    marshaller_context: MarshallerContext = field(default_factory=get_default_context)
    schema_context: SchemaContext = field(default_factory=get_default_schema_context)
    validate_output: bool = True
    allow_unsigned_auth: bool = True
    priority: int = 50
    event_handler_type: Type[EventHandlerABC] = EventHandler

    def create(self, action: Action) -> EventHandlerABC:
        fn, auth_kwarg_name = separate_auth_kwarg(action.fn)
        authorizer = None
        auth_marshaller = None
        if auth_kwarg_name or action.access_control != ALLOW_ALL:
            authorizer = get_default_authorizer()
            if self.allow_unsigned_auth:
                auth_marshaller = self.marshaller_context.get_marshaller(Authorization)
        param_marshaller = get_marshaller_for_params(fn, set(), self.marshaller_context)
        param_schema = get_schema_for_params(fn, set(), self.schema_context)
        sig = inspect.signature(fn)
        result_marshaller = None
        result_schema = None
        if sig.return_annotation != inspect.Signature.empty:
            result_marshaller = self.marshaller_context.get_marshaller(
                sig.return_annotation
            )
            if self.validate_output:
                result_schema = self.schema_context.schema_from_type(
                    resolve_forward_refs(sig.return_annotation)
                )
        return self.event_handler_type(
            action=action,
            param_marshaller=param_marshaller,
            param_schema=param_schema,
            result_marshaller=result_marshaller,
            result_schema=result_schema,
            auth_kwarg_name=auth_kwarg_name,
            auth_marshaller=auth_marshaller,
            authorizer=authorizer,
            priority=self.priority,
        )


def separate_auth_kwarg(fn: Callable) -> Tuple[Callable, Optional[str]]:
    attr_name = get_inject_at(fn)
    if not attr_name:
        return fn, attr_name

    def noop(**_):
        """This function exists to transfer signature data. it is never called"""

    sig = inspect.signature(fn)
    sig = sig.replace(
        parameters=[p for p in sig.parameters.values() if p.name != attr_name]
    )
    noop.__signature__ = sig

    return noop, attr_name



"""
This module sets up a celery app using any default Action with a FixedRateTrigger
to be run from a celery worker.
All such actions are added to the global namespace as celery tasks.
Celery supports use of rabbitmq, redis, and even sqs.

Example:
    This example contains 3 processes:
        * A rabbitmq message broker process
        * A celery worker process
        * A celery beat scheduler process

    Rabbitmq is available through docker using `docker run -d -p 5672:5672 rabbitmq`. It should start automatically
    when downloaded (You can also restart it in the docker console)

    In a terminal, start a celery worker process with :
        `celery -A servey.servey_celery.celery_app worker --loglevel=INFO`

    In a second terminal, start a celery beat scheduler with:
        `celery -A servey.servey_celery.celery_app beat --loglevel=INFO`

    Log output for scheduled tasks should appear in the output for the worker.

"""
import logging
import os
from celery import Celery

from marshy.factory.impl_marshaller_factory import get_impls
from servey.servey_celery.celery_config.celery_config_abc import CeleryConfigABC

# Setup app and tasks...
_CELERY_BROKER = os.environ.get("CELERY_BROKER")
_LOGGER = logging.getLogger(__name__)
_LOGGER.info(f"Starting celery with broker {_CELERY_BROKER}")
app = Celery(broker=_CELERY_BROKER)
for impl in get_impls(CeleryConfigABC):
    impl().configure(app, globals())



"""
Module for running a local python debugging server based on Uvicorn, along with a
local_schedule_mount for actions.
"""
import argparse
import json
import logging
import os

from servey.errors import ServeyError

LOGLEVEL = os.environ.get("LOGLEVEL", "INFO").upper()
logging.basicConfig(level=LOGLEVEL)
DEBUG = int(os.environ.get("SERVER_DEBUG", "1")) == 1
HOST = os.environ.get("SERVER_HOST", "0.0.0.0")
PORT = int(os.environ.get("SERVER_PORT", "8000"))
CELERY_BROKER = os.environ.get("CELERY_BROKER")


def start_http_server():
    import uvicorn

    app = "servey.servey_starlette.starlette_app:app"
    uvicorn.run(
        app,
        host=HOST,
        port=PORT,
        log_level=LOGLEVEL.lower(),
        reload=DEBUG,
        reload_includes=["*.j2", "*.py"],
        reload_dirs=["."],
    )


# pylint: disable=W0611
def start_scheduler():
    if CELERY_BROKER:
        # noinspection PyUnresolvedReferences
        import servey.servey_celery.celery_app
    else:
        os.environ["SERVEY_DAEMON"] = "1"
        # noinspection PyUnresolvedReferences
        import servey.servey_thread.__main__


def generate_openapi_schema():
    if "SERVEY_STRICT_SCHEMA" not in os.environ:
        os.environ["SERVEY_STRICT_SCHEMA"] = "1"
    from servey.servey_starlette.route_factory.openapi_route_factory import (
        OpenapiRouteFactory,
    )

    factory = OpenapiRouteFactory()
    schema = factory.openapi_schema()
    output_file = os.environ.get("OUTPUT_FILE", "openapi.json")
    json.dump(schema, output_file)


def generate_graphql_schema():
    from servey.servey_aws.serverless.yml_config.appsync_config import AppsyncConfig

    config = AppsyncConfig()
    config.build_appsync_schema()


def main():
    parser = argparse.ArgumentParser(description="Servey")
    parser.add_argument("--run", default="server")
    args, _ = parser.parse_known_args()
    if args.run == "sls":
        # noinspection PyUnresolvedReferences
        from servey.servey_aws.serverless.__main__ import main as sls_main

        sls_main()
    elif args.run == "openapi":
        generate_openapi_schema()
    elif args.run == "graphql-schema":
        generate_graphql_schema()
    elif args.run == "action":
        # noinspection PyUnresolvedReferences
        from servey.servey_direct.__main__ import main as direct_main

        direct_main()
    elif args.run == "server":
        start_scheduler()
        start_http_server()
    else:
        raise ServeyError(f"unknown:{args.run}")


if __name__ == "__main__":
    main()



import json
import os
from dataclasses import dataclass, field
from typing import List, Optional, Any, Dict

import boto3
from boto3.dynamodb.conditions import Key
from marshy import get_default_context
from marshy.marshaller.marshaller_abc import MarshallerABC

from servey.security.access_control.allow_none import ALLOW_NONE
from servey.security.authorization import Authorization
from servey.servey_aws import is_lambda_env
from servey.subscription.subscription import Subscription, T
from servey.subscription.subscription_service import (
    SubscriptionServiceFactoryABC,
    SubscriptionServiceABC,
)


@dataclass
class AWSWebsocketSubscriptionService(SubscriptionServiceABC):
    connection_table: Any = field(
        default_factory=lambda: boto3.resource("dynamodb").Table(
            os.environ["CONNECTION_TABLE_NAME"]
        )
    )
    apis_by_endpoint_url: Dict[str, Any] = field(default_factory=dict)
    authorization_marshaller: MarshallerABC[Authorization] = field(
        default_factory=lambda: get_default_context().get_marshaller(
            Optional[Authorization]
        )
    )

    def get_api_for_endpoint(self, endpoint_url: str):
        api = self.apis_by_endpoint_url.get(endpoint_url)
        if not api:
            api = boto3.client("apigatewaymanagementapi", endpoint_url=endpoint_url)
            self.apis_by_endpoint_url[endpoint_url] = api
        return api

    def publish(self, subscription: Subscription[T], event: T):
        data = json.dumps(subscription.event_marshaller.dump(event)).encode("utf-8")
        kwargs = {
            "Select": "SPECIFIC_ATTRIBUTES",
            "ProjectionExpression": "connection_id,endpoint_url,user_authorization",
            "IndexName": "gsi__subscription_name__connection_id",
            "KeyConditionExpression": Key("subscription_name").eq(subscription.name),
        }
        while True:
            response = self.connection_table.query(**kwargs)
            items = response.get("Items") or []
            for item in items:
                api = self.get_api_for_endpoint(item["endpoint_url"])
                if subscription.event_filter:
                    user_authorization = item.get("user_authorization")
                    if user_authorization:
                        user_authorization = self.authorization_marshaller.load(
                            user_authorization
                        )
                    if not subscription.event_filter.should_publish(
                        event, user_authorization
                    ):
                        continue
                api.post_to_connection(Data=data, ConnectionId=item["connection_id"])
            kwargs["ExclusiveStartKey"] = response.get("LastEvaluatedKey")
            if not kwargs["ExclusiveStartKey"]:
                break


class AWSWebsocketSubscriptionServiceFactory(SubscriptionServiceFactoryABC):
    def create(
        self, subscriptions: List[Subscription]
    ) -> Optional[SubscriptionServiceABC]:
        if not is_lambda_env():
            return
        subscribable = [s for s in subscriptions if s.access_control != ALLOW_NONE]
        if subscribable:
            return AWSWebsocketSubscriptionService()



from abc import ABC, abstractmethod
from typing import Tuple

from marshy.factory.impl_marshaller_factory import get_impls
from marshy.types import ExternalType

from servey.action.action import Action
from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.servey_aws.event_handler.event_handler_abc import EventHandlerABC
from servey.trigger.web_trigger import WebTrigger


class RouterABC(ABC):
    """A router will attempt to create an event handler given a specific lambda event"""

    priority: int = 100

    @abstractmethod
    def create_handler(self, event: ExternalType, context) -> EventHandlerABC:
        """Create a handler for the event given"""

    @property
    def web_trigger_actions(self) -> Tuple[Tuple[Action, WebTrigger], ...]:
        web_trigger_actions = getattr(self, "_web_trigger_actions", None)
        if web_trigger_actions is None:
            web_trigger_actions = list(find_actions_with_trigger_type(WebTrigger))
            web_trigger_actions.sort(
                key=lambda n: len(n[1].path or f"/actions/{n[0].name}"), reverse=True
            )
            web_trigger_actions = tuple(web_trigger_actions)
            setattr(self, "_web_trigger_actions", web_trigger_actions)
        return web_trigger_actions


def find_routers() -> Tuple[RouterABC]:
    routers = [router() for router in get_impls(RouterABC)]
    routers.sort(key=lambda r: r.priority, reverse=True)
    return tuple(routers)



from dataclasses import dataclass, field
from typing import Type

from strawberry import LazyType

from servey.servey_strawberry.schema_factory import SchemaFactory


@dataclass(frozen=True)
class SchemaFactoryLazyType(LazyType):
    """
    This is a bit of a hack - pretend to be a standard LazyType to get strawberry to resolve the type
    """

    schema_factory: SchemaFactory = field(repr=False, hash=False, default=None)

    def resolve_type(self) -> Type:
        return self.schema_factory.types.get(self.type_name)

    def __call__(self):  # pragma: no cover
        # this empty call method allows SchemaFactoryLazyInput to be used in generic types
        return None

    @property
    def __name__(self):
        return self.type_name



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator

from starlette.routing import Route, Mount
from starlette.staticfiles import StaticFiles

from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC


@dataclass
class StaticSiteRouteFactory(RouteFactoryABC):
    priority: int = 80
    path: str = "/"
    directory: Path = Path(os.environ.get("SERVEY_STATIC_SITE_DIR") or "static_site")

    def create_routes(self) -> Iterator[Route]:
        if self.directory.exists():
            yield Mount(
                self.path,
                app=StaticFiles(directory=self.directory, html=True),
                name="static_site",
            )



from abc import abstractmethod, ABC
from typing import Optional

from marshy.factory.impl_marshaller_factory import get_impls

from servey.security.authorizer.authorizer_abc import AuthorizerABC


class AuthorizerFactoryABC(ABC):
    priority: int = 100

    @abstractmethod
    def create_authorizer(self) -> Optional[AuthorizerABC]:
        """Create a new authorizer instance. Return None if this was not possible"""


def create_authorizer():
    """
    Use the highest priority factory that can actually do it to create an authorizer.
    For example, the KmsAuthorizerFactory requires that boto3 is available and that a KMS_KEY_ID specified
    as an environment variable, so will yield nothing if either of these constraints are not met.
    The JwtAuthorizationFactory has lower priority, will make up a key on the fly as required, but requires
    that pyjwt is installed.
    """
    factories = list(get_impls(AuthorizerFactoryABC))
    factories.sort(key=lambda f: f.priority, reverse=True)
    for factory in factories:
        authorizer = factory().create_authorizer()
        if authorizer:
            return authorizer


_default_authorizer = None


# pylint: disable=W0603
def get_default_authorizer():
    global _default_authorizer
    if not _default_authorizer:
        _default_authorizer = create_authorizer()
    return _default_authorizer



from dataclasses import dataclass
from enum import Enum
from typing import Optional

from servey.trigger.trigger_abc import TriggerABC


class WebTriggerMethod(Enum):
    DELETE = "delete"
    GET = "get"
    HEAD = "head"
    OPTIONS = "options"
    PATCH = "patch"
    POST = "post"
    PUT = "put"


@dataclass(frozen=True)
class WebTrigger(TriggerABC):
    method: WebTriggerMethod = WebTriggerMethod.POST
    path: Optional[str] = None


WEB_GET = WebTrigger(WebTriggerMethod.GET)
WEB_POST = WebTrigger(WebTriggerMethod.POST)
BODY_METHODS = (
    WebTriggerMethod.PATCH,
    WebTriggerMethod.POST,
    WebTriggerMethod.PUT,
)
UPDATE_METHODS = (
    WebTriggerMethod.DELETE,
    WebTriggerMethod.PATCH,
    WebTriggerMethod.POST,
    WebTriggerMethod.PUT,
)



import os

from dataclasses import dataclass, field
from logging import getLogger
from typing import Iterator

from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Route

from servey.security.authenticator.password_authenticator_abc import (
    PasswordAuthenticatorABC,
    get_default_password_authenticator,
)
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC

LOGGER = getLogger(__name__)


class AuthenticatorRouteFactory(RouteFactoryABC):
    """Route factory which performs authentication."""

    def create_routes(self) -> Iterator[Route]:
        # Create an authenticator object based on username and password
        if os.environ.get("SERVEY_DEBUG_AUTHENTICATOR") == "0":
            return
        authenticator = PasswordLoginEndpoint()
        yield Route(path="/login", methods=["post"], endpoint=authenticator.login)


@dataclass
class PasswordLoginEndpoint:
    password_authenticator: PasswordAuthenticatorABC = field(
        default_factory=get_default_password_authenticator
    )
    authorizer: AuthorizerABC = field(default_factory=get_default_authorizer)

    async def login(self, request: Request) -> JSONResponse:
        form_data = await request.form()
        authorization = self.password_authenticator.authenticate(
            form_data["username"], form_data["password"]
        )
        if authorization:
            return JSONResponse(
                {
                    "access_token": self.authorizer.encode(authorization),
                    "token_type": "bearer",
                }
            )
        raise HTTPException(status_code=400, detail="Incorrect username or password")



"""
This package implements the MVC paradigm using servey actions and jinja, allowing production of web pages.
It is an optional extra. The hosted version uses starlette, and the AWS version uses API gateway.
"""



from typing import List, Optional

from servey.subscription.subscription import Subscription, T
from servey.subscription.subscription_service import (
    SubscriptionServiceABC,
    SubscriptionServiceFactoryABC,
)


class CelerySubscriptionService(SubscriptionServiceABC):
    """
    Service designed to work with celery to cluster subscriptions.
    """

    def publish(self, subscription: Subscription[T], event: T):
        from servey.servey_celery import celery_app

        for action in subscription.action_subscribers:
            task = getattr(celery_app, action.name)
            task.delay(*[event])


class CelerySubscriptionServiceFactory(SubscriptionServiceFactoryABC):
    """
    Factory which sets up a celery subscription service only if a celery
    broker was specified
    """

    def create(
        self, subscriptions: List[Subscription]
    ) -> Optional[SubscriptionServiceABC]:
        from servey.servey_celery import has_celery_broker

        if has_celery_broker():
            return CelerySubscriptionService()



import os
from pathlib import Path

from marshy.types import ExternalItemType

from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.servey_aws.serverless.yml_config.yml_config_abc import (
    YmlConfigABC,
    ensure_ref_in_file,
    create_yml_file,
)


class CloudfrontConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    static_site_directory: Path = Path(
        os.environ.get("SERVEY_STATIC_SITE_DIR") or "static_site"
    )
    static_site_bucket_resource_yml_file: str = (
        "serverless_servey/cloudfront_resource.yml"
    )

    def configure(self, main_serverless_yml_file: str):
        has_static_site = self.static_site_directory.exists()
        has_web_page = False
        try:
            # noinspection PyUnresolvedReferences
            from servey.servey_web_page.web_page_trigger import WebPageTrigger

            has_web_page = next(
                (True for _ in find_actions_with_trigger_type(WebPageTrigger)), False
            )
        except ImportError:
            pass
        if not has_static_site and not has_web_page:
            return
        # If there is a static site
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["resources"],
            self.static_site_bucket_resource_yml_file,
        )
        cloudfront_resource_yml = self.build_cloudfront_resource_yml()
        create_yml_file(
            self.static_site_bucket_resource_yml_file, cloudfront_resource_yml
        )

    @staticmethod
    def build_cloudfront_resource_yml() -> ExternalItemType:
        static_site_bucket_resource = {
            "Resources": {
                "End2EndIndexFn": {
                    "Type": "AWS::CloudFront::Function",
                    "Properties": {
                        "Name": "End2EndIndexFn",
                        "AutoPublish": True,
                        "FunctionConfig": {
                            "Comment": "Add index.html when path ends with /",
                            "Runtime": "cloudfront-js-1.0",
                        },
                        "FunctionCode": "\n".join(
                            (
                                "function handler(event) {",
                                "    var request = event.request;",
                                "    var uri = request.uri;",
                                "",
                                "    if (uri.endsWith('/')) {",
                                "        request.uri += 'index.html';",
                                "    }",
                                "    else if (!uri.includes('.')) {",
                                "        request.uri += '/index.html';",
                                "    }",
                                "",
                                "    return request;",
                                "}",
                                "",
                            )
                        ),
                    },
                },
                "End2EndDistribution": {
                    "Type": "AWS::CloudFront::Distribution",
                    "Properties": {
                        "DistributionConfig": {
                            "Enabled": "true",
                            "DefaultRootObject": "/",
                            "DefaultCacheBehavior": {
                                "AllowedMethods": ["GET", "HEAD"],
                                "MinTTL": "0",
                                "MaxTTL": "0",
                                "DefaultTTL": "0",
                                "TargetOriginId": "s3Origin",
                                "ForwardedValues": {"QueryString": "false"},
                                "ViewerProtocolPolicy": "redirect-to-https",
                                "FunctionAssociations": [
                                    {
                                        "EventType": "viewer-request",
                                        "FunctionARN": {"Ref": "End2EndIndexFn"},
                                    }
                                ],
                            },
                            "CacheBehaviors": [
                                {
                                    "AllowedMethods": [
                                        "HEAD",
                                        "DELETE",
                                        "POST",
                                        "GET",
                                        "OPTIONS",
                                        "PUT",
                                        "PATCH",
                                    ],
                                    "TargetOriginId": "apiGatewayAPIOrigin",
                                    "ForwardedValues": {
                                        "QueryString": True,
                                        "Cookies": {"Forward": "all"},
                                    },
                                    "ViewerProtocolPolicy": "https-only",
                                    "MinTTL": "0",
                                    "MaxTTL": "6",
                                    "DefaultTTL": "3",
                                    "PathPattern": "actions/*",
                                },
                                {
                                    "AllowedMethods": [
                                        "HEAD",
                                        "DELETE",
                                        "POST",
                                        "GET",
                                        "OPTIONS",
                                        "PUT",
                                        "PATCH",
                                    ],
                                    "TargetOriginId": "appsyncAPIOrigin",
                                    "ForwardedValues": {
                                        "QueryString": False,
                                        "Cookies": {"Forward": "all"},
                                    },
                                    "ViewerProtocolPolicy": "https-only",
                                    "MinTTL": "0",
                                    "MaxTTL": "6",
                                    "DefaultTTL": "3",
                                    "PathPattern": "graphql",
                                },
                            ],
                            "Origins": [
                                {
                                    "DomainName": {
                                        "Fn::Join": [
                                            "",
                                            [
                                                {"Ref": "ApiGatewayRestApi"},
                                                ".execute-api.${self:provider.region}.amazonaws.com",
                                            ],
                                        ]
                                    },
                                    "Id": "apiGatewayAPIOrigin",
                                    "OriginPath": "/${self:provider.stage}",
                                    "CustomOriginConfig": {
                                        "OriginProtocolPolicy": "https-only"
                                    },
                                },
                                {
                                    "DomainName": {
                                        "Fn::Select": [
                                            0,
                                            {
                                                "Fn::Split": [
                                                    "/graphql",
                                                    {
                                                        "Fn::Select": [
                                                            1,
                                                            {
                                                                "Fn::Split": [
                                                                    "https://",
                                                                    {
                                                                        "Fn::GetAtt": [
                                                                            "GraphQlApi",
                                                                            "GraphQLUrl",
                                                                        ]
                                                                    },
                                                                ]
                                                            },
                                                        ]
                                                    },
                                                ]
                                            },
                                        ]
                                    },
                                    "Id": "appsyncAPIOrigin",
                                    "CustomOriginConfig": {
                                        "OriginProtocolPolicy": "https-only"
                                    },
                                },
                                {
                                    "DomainName": "${self:custom.staticBucket}.s3.amazonaws.com",
                                    "Id": "s3Origin",
                                    "S3OriginConfig": {
                                        "OriginAccessIdentity": {
                                            "Fn::Join": [
                                                "",
                                                [
                                                    "origin-access-identity/cloudfront/",
                                                    {"Ref": "OriginAccessIdentity"},
                                                ],
                                            ]
                                        }
                                    },
                                },
                            ],
                        }
                    },
                },
            }
        }
        return static_site_bucket_resource



from abc import ABC, abstractmethod
from typing import Optional

from marshy.factory.impl_marshaller_factory import get_impls

from servey.security.authorization import Authorization


class PasswordAuthenticatorABC(ABC):
    priority: int = 100

    @abstractmethod
    def authenticate(self, username: str, password: str) -> Optional[Authorization]:
        """Authenticate the username and password given"""


def get_default_password_authenticator() -> PasswordAuthenticatorABC:
    impls = list(get_impls(PasswordAuthenticatorABC))
    impls.sort(key=lambda i: i.priority, reverse=True)
    return impls[0]()



import logging

from marshy.factory.impl_marshaller_factory import register_impl
from marshy.marshaller_context import MarshallerContext

from servey.subscription.subscription_service import SubscriptionServiceFactoryABC

priority = 90
LOGGER = logging.getLogger(__name__)


def configure(context: MarshallerContext):
    from marshy.factory.dataclass_marshaller_factory import DataclassMarshallerFactory
    from servey.util.to_second_datetime_marshaller import (
        ToSecondDatetimeMarshaller,
    )

    context.register_factory(
        DataclassMarshallerFactory(priority=101, exclude_dumped_values=tuple())
    )
    context.register_marshaller(ToSecondDatetimeMarshaller())
    configure_finders(context)
    configure_asyncio_subscriptions(context)
    configure_auth(context)
    configure_starlette(context)
    configure_aws(context)
    configure_serverless(context)
    configure_strawberry(context)
    configure_strawberry_starlette(context)
    configure_celery(context)
    configure_jinja2(context)


def configure_finders(context: MarshallerContext):
    from servey.finder.action_finder_abc import ActionFinderABC
    from servey.finder.module_action_finder import ModuleActionFinder
    from servey.finder.module_subscription_finder import ModuleSubscriptionFinder
    from servey.finder.subscription_finder_abc import SubscriptionFinderABC

    register_impl(ActionFinderABC, ModuleActionFinder, context)
    register_impl(SubscriptionFinderABC, ModuleSubscriptionFinder, context)


def configure_asyncio_subscriptions(context: MarshallerContext):
    from servey.servey_thread.asyncio_subscription_service import (
        AsyncioSubscriptionServiceFactory,
    )

    register_impl(
        SubscriptionServiceFactoryABC, AsyncioSubscriptionServiceFactory, context
    )


def configure_auth(context: MarshallerContext):
    from servey.security.authorizer.authorizer_factory_abc import AuthorizerFactoryABC
    from servey.security.authorizer.jwt_authorizer_factory import JwtAuthorizerFactory
    from servey.security.authenticator.password_authenticator_abc import (
        PasswordAuthenticatorABC,
    )
    from servey.security.authenticator.root_password_authenticator import (
        RootPasswordAuthenticator,
    )

    register_impl(AuthorizerFactoryABC, JwtAuthorizerFactory, context)
    register_impl(PasswordAuthenticatorABC, RootPasswordAuthenticator, context)


def configure_starlette(context: MarshallerContext):
    try:
        configure_starlette_action_endpoint_factory(context)
        configure_starlette_route_factory(context)
    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_starlette_action_endpoint_factory(context: MarshallerContext):
    from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
        ActionEndpointFactoryABC,
    )
    from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory import (
        ActionEndpointFactory,
    )
    from servey.servey_starlette.action_endpoint.factory.authorizing_action_endpoint_factory import (
        AuthorizingActionEndpointFactory,
    )
    from servey.servey_starlette.action_endpoint.factory.caching_action_endpoint_factory import (
        CachingActionEndpointFactory,
    )
    from servey.servey_starlette.action_endpoint.factory.self_action_endpoint_factory import (
        SelfActionEndpointFactory,
    )

    register_impl(ActionEndpointFactoryABC, ActionEndpointFactory, context)
    register_impl(ActionEndpointFactoryABC, AuthorizingActionEndpointFactory, context)
    register_impl(ActionEndpointFactoryABC, CachingActionEndpointFactory, context)
    register_impl(ActionEndpointFactoryABC, SelfActionEndpointFactory, context)


def configure_starlette_route_factory(context: MarshallerContext):
    from servey.servey_starlette.route_factory.action_route_factory import (
        ActionRouteFactory,
    )
    from servey.servey_starlette.route_factory.authenticator_route_factory import (
        AuthenticatorRouteFactory,
    )
    from servey.servey_starlette.route_factory.openapi_route_factory import (
        OpenapiRouteFactory,
    )
    from servey.servey_starlette.route_factory.route_factory_abc import RouteFactoryABC
    from servey.servey_starlette.route_factory.subscription_route_factory import (
        SubscriptionRouteFactory,
    )
    from servey.servey_starlette.route_factory.asyncapi_route_factory import (
        AsyncapiRouteFactory,
    )
    from servey.servey_starlette.route_factory.static_site_route_factory import (
        StaticSiteRouteFactory,
    )

    register_impl(RouteFactoryABC, ActionRouteFactory, context)
    register_impl(RouteFactoryABC, AuthenticatorRouteFactory, context)
    register_impl(RouteFactoryABC, OpenapiRouteFactory, context)
    register_impl(RouteFactoryABC, SubscriptionRouteFactory, context)
    register_impl(SubscriptionServiceFactoryABC, SubscriptionRouteFactory, context)
    register_impl(RouteFactoryABC, AsyncapiRouteFactory, context)
    register_impl(RouteFactoryABC, StaticSiteRouteFactory, context)


def configure_strawberry(context: MarshallerContext):
    try:
        from servey.servey_strawberry.handler_filter.handler_filter_abc import (
            HandlerFilterABC,
        )
        from servey.servey_strawberry.handler_filter.authorization_handler_filter import (
            AuthorizationHandlerFilter,
        )
        from servey.servey_strawberry.handler_filter.strawberry_type_handler_filter import (
            StrawberryTypeHandlerFilter,
        )

        from servey.servey_strawberry.entity_factory.entity_factory_abc import (
            EntityFactoryABC,
        )
        from servey.servey_strawberry.entity_factory.dataclass_factory import (
            DataclassFactory,
        )
        from servey.servey_strawberry.entity_factory.enum_factory import (
            EnumFactory,
        )
        from servey.servey_strawberry.entity_factory.forward_ref_factory import (
            ForwardRefFactory,
        )
        from servey.servey_strawberry.entity_factory.generic_factory import (
            GenericFactory,
        )
        from servey.servey_strawberry.entity_factory.no_op_factory import (
            NoOpFactory,
        )

        register_impl(HandlerFilterABC, AuthorizationHandlerFilter, context)
        register_impl(HandlerFilterABC, StrawberryTypeHandlerFilter, context)

        register_impl(EntityFactoryABC, DataclassFactory, context)
        register_impl(EntityFactoryABC, EnumFactory, context)
        register_impl(EntityFactoryABC, ForwardRefFactory, context)
        register_impl(EntityFactoryABC, GenericFactory, context)
        register_impl(EntityFactoryABC, NoOpFactory, context)

    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_strawberry_starlette(context: MarshallerContext):
    try:
        from servey.servey_starlette.route_factory.route_factory_abc import (
            RouteFactoryABC,
        )
        from servey.servey_strawberry.strawberry_starlette_route_factory import (
            StrawberryStarletteRouteFactory,
        )

        register_impl(RouteFactoryABC, StrawberryStarletteRouteFactory, context)

    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_aws(context: MarshallerContext):
    try:
        from servey.servey_aws.authorizer.kms_authorizer_factory import (
            KmsAuthorizerFactory,
        )
        from servey.security.authorizer.authorizer_factory_abc import (
            AuthorizerFactoryABC,
        )

        register_impl(AuthorizerFactoryABC, KmsAuthorizerFactory, context)

        from servey.servey_aws.event_handler.event_handler_abc import (
            EventHandlerFactoryABC,
        )
        from servey.servey_aws.event_handler.event_handler import EventHandlerFactory
        from servey.servey_aws.event_handler.api_gateway_event_handler import (
            ApiGatewayEventHandlerFactory,
        )
        from servey.servey_aws.event_handler.appsync_event_handler import (
            AppsyncEventHandlerFactory,
        )
        from servey.servey_aws.event_handler.sqs_event_handler import (
            SqsEventHandlerFactory,
        )

        register_impl(EventHandlerFactoryABC, EventHandlerFactory, context)
        register_impl(EventHandlerFactoryABC, AppsyncEventHandlerFactory, context)
        register_impl(EventHandlerFactoryABC, ApiGatewayEventHandlerFactory, context)
        register_impl(EventHandlerFactoryABC, SqsEventHandlerFactory, context)

        from servey.servey_aws.websocket_subscription_service import (
            AWSWebsocketSubscriptionServiceFactory,
        )

        register_impl(
            SubscriptionServiceFactoryABC,
            AWSWebsocketSubscriptionServiceFactory,
            context,
        )

        from servey.servey_aws.sqs_subscription_service import (
            SqsSubscriptionServiceFactory,
        )

        register_impl(
            SubscriptionServiceFactoryABC,
            SqsSubscriptionServiceFactory,
            context,
        )

        from servey.servey_aws.router.router_abc import RouterABC
        from servey.servey_aws.router.api_gateway_router import APIGatewayRouter
        from servey.servey_aws.router.appsync_router import AppsyncRouter
        from servey.servey_aws.router.router import Router

        register_impl(RouterABC, APIGatewayRouter, context)
        register_impl(RouterABC, AppsyncRouter, context)
        register_impl(RouterABC, Router, context)

    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_serverless(context: MarshallerContext):
    try:
        from servey.servey_aws.serverless.yml_config.yml_config_abc import YmlConfigABC
        from servey.servey_aws.serverless.yml_config.action_function_config import (
            ActionFunctionConfig,
        )

        # pylint: disable=E0001
        from servey.servey_aws.serverless.yml_config.appsync_config import AppsyncConfig
        from servey.servey_aws.serverless.yml_config.kms_key_config import KmsKeyConfig
        from servey.servey_aws.serverless.yml_config.subscription_function_config import (
            SubscriptionFunctionConfig,
        )

        register_impl(YmlConfigABC, ActionFunctionConfig, context)
        register_impl(YmlConfigABC, AppsyncConfig, context)
        register_impl(YmlConfigABC, KmsKeyConfig, context)
        register_impl(YmlConfigABC, SubscriptionFunctionConfig, context)

        from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
            TriggerHandlerABC,
        )
        from servey.servey_aws.serverless.trigger_handler.web_trigger_handler import (
            WebTriggerHandler,
        )
        from servey.servey_aws.serverless.trigger_handler.fixed_rate_trigger_handler import (
            FixedRateTriggerHandler,
        )

        register_impl(TriggerHandlerABC, WebTriggerHandler, context)
        register_impl(TriggerHandlerABC, FixedRateTriggerHandler, context)

        from servey.servey_aws.serverless.yml_config.cloudfront_config import (
            CloudfrontConfig,
        )
        from servey.servey_aws.serverless.yml_config.static_site_bucket_config import (
            StaticSiteBucketConfig,
        )

        register_impl(YmlConfigABC, CloudfrontConfig, context)
        register_impl(YmlConfigABC, StaticSiteBucketConfig, context)

    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_celery(context: MarshallerContext):
    try:
        from servey.servey_celery.celery_subscription_service import (
            CelerySubscriptionServiceFactory,
        )

        register_impl(
            SubscriptionServiceFactoryABC, CelerySubscriptionServiceFactory, context
        )

        from servey.servey_celery.celery_config.celery_config_abc import CeleryConfigABC
        from servey.servey_celery.celery_config.fixed_rate_trigger_config import (
            FixedRateTriggerConfig,
        )
        from servey.servey_celery.celery_config.subscription_config import (
            SubscriptionConfig,
        )

        register_impl(CeleryConfigABC, FixedRateTriggerConfig, context)
        register_impl(CeleryConfigABC, SubscriptionConfig, context)

    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_jinja2(context: MarshallerContext):
    try:
        # We import Template to test that jinja2 is actually present
        # pylint: disable=W0611
        from jinja2 import Template

        configure_web_page_action_endpoint_factory(context)
        configure_web_page_event_handler(context)
        configure_web_page_trigger_handler(context)
    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_web_page_action_endpoint_factory(context: MarshallerContext):
    try:
        from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
            ActionEndpointFactoryABC,
        )
        from servey.servey_web_page.web_page_action_endpoint_factory import (
            WebPageActionEndpointFactory,
        )

        register_impl(ActionEndpointFactoryABC, WebPageActionEndpointFactory, context)
    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_web_page_event_handler(context: MarshallerContext):
    try:
        from servey.servey_aws.event_handler.event_handler_abc import (
            EventHandlerFactoryABC,
        )
        from servey.servey_web_page.web_page_event_handler import (
            WebPageEventHandlerFactory,
        )

        register_impl(EventHandlerFactoryABC, WebPageEventHandlerFactory, context)
    except ModuleNotFoundError as e:
        raise_non_ignored(e)


def configure_web_page_trigger_handler(context: MarshallerContext):
    try:
        from servey.servey_aws.serverless.trigger_handler.trigger_handler_abc import (
            TriggerHandlerABC,
        )
        from servey.servey_web_page.web_page_trigger_handler import (
            WebPageTriggerHandler,
        )

        register_impl(TriggerHandlerABC, WebPageTriggerHandler, context)
    except ModuleNotFoundError as e:
        raise_non_ignored(e)


# Due to the use of extras, certain modules may not be present, but that's okay
_NO_MODULE_NAMED = "No module named '"
_IGNORABLE_MISSING_MODULE_NAMES = {
    "celery",
    "requests",
    "jinja2",
    "ruamel",
    "ruamel.yaml",
    "boto3",
    "starlette",
    "strawberry",
}


def raise_non_ignored(e: ModuleNotFoundError):
    msg = str(e)
    if msg.startswith(_NO_MODULE_NAMED):
        module_name = msg[len(_NO_MODULE_NAMED) : -1]
        if module_name in _IGNORABLE_MISSING_MODULE_NAMES:
            LOGGER.debug(msg)
            return
    raise e



from dataclasses import dataclass, field
from typing import List, Optional, Set

from servey.action.action import Action
from servey.security.access_control.allow_all import ALLOW_ALL
from servey.security.authorization import get_inject_at
from servey.security.authorizer.authorizer_abc import AuthorizerABC
from servey.security.authorizer.authorizer_factory_abc import get_default_authorizer
from servey.servey_starlette.action_endpoint.action_endpoint_abc import (
    ActionEndpointABC,
)
from servey.servey_starlette.action_endpoint.authorizing_action_endpoint import (
    AuthorizingActionEndpoint,
)
from servey.servey_starlette.action_endpoint.factory.action_endpoint_factory_abc import (
    ActionEndpointFactoryABC,
)


@dataclass
class AuthorizingActionEndpointFactory(ActionEndpointFactoryABC):
    priority: int = 200
    authorizer: AuthorizerABC = field(default_factory=get_default_authorizer)
    skip: bool = False

    def create(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        if self.skip:
            return

        auth_kwarg_name = get_inject_at(action.fn)
        if not auth_kwarg_name and action.access_control == ALLOW_ALL:
            return
        if auth_kwarg_name:
            skip_args = skip_args.copy()
            skip_args.add(auth_kwarg_name)
        action_endpoint = self._get_wrapped_endpoint(action, skip_args, factories)
        if action_endpoint:
            return AuthorizingActionEndpoint(
                action_endpoint=action_endpoint,
                authorizer=self.authorizer,
                auth_kwarg_name=auth_kwarg_name,
            )

    def _get_wrapped_endpoint(
        self,
        action: Action,
        skip_args: Set[str],
        factories: List[ActionEndpointFactoryABC],
    ) -> Optional[ActionEndpointABC]:
        self.skip = True
        try:
            for factory in factories:
                action_endpoint = factory.create(action, skip_args, factories)
                if action_endpoint:
                    return action_endpoint
        finally:
            self.skip = False



import dataclasses
import inspect
from dataclasses import is_dataclass, fields, dataclass, MISSING
from datetime import datetime
from decimal import Decimal
from typing import Type, Optional, Dict, Any, Callable, Union

import strawberry
from dateutil.relativedelta import relativedelta
from strawberry.dataloader import DataLoader, AbstractCache

from servey.action.action import Action, get_action
from servey.action.batch_invoker import BatchInvoker
from servey.cache_control.cache_control_abc import CacheControlABC
from servey.servey_strawberry.entity_factory.entity_factory_abc import (
    EntityFactoryABC,
)
from servey.servey_strawberry.schema_factory import SchemaFactory
from servey.servey_strawberry.schema_factory_lazy_input import SchemaFactoryLazyInput
from servey.servey_strawberry.schema_factory_lazy_type import SchemaFactoryLazyType


class DataclassFactory(EntityFactoryABC):
    priority: int = 120

    def create_type(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if not is_dataclass(annotation):
            return
        # noinspection PyTypeChecker
        schema_factory.types[annotation.__name__] = SchemaFactoryLazyType(
            type_name=annotation.__name__, module="", schema_factory=schema_factory
        )

        annotations = {}
        params = {"__annotations__": annotations}
        # noinspection PyDataclass
        for f in fields(annotation):
            annotations[f.name] = schema_factory.get_type(f.type)

        # Check for functions decorated with @action
        for key, value in annotation.__dict__.items():
            resolvable = get_action(value)
            if resolvable:
                build_resolvable_field(resolvable, schema_factory, key, params)

        wrap_type = type(annotation.__name__, tuple(), params)
        type_ = strawberry.type(wrap_type)
        schema_factory.types[annotation.__name__] = type_
        # noinspection PyDataclass
        for field in fields(type_):
            field.type = schema_factory.get_type(field.type)
        return type_

    def create_input(
        self, annotation: Type, schema_factory: SchemaFactory
    ) -> Optional[Type]:
        if not is_dataclass(annotation):
            return
        name = f"{annotation.__name__}Input"
        input_ = schema_factory.inputs.get(name)
        if input_:
            return input_
        # noinspection PyTypeChecker
        schema_factory.inputs[name] = SchemaFactoryLazyInput(
            type_name=name, module="", schema_factory=schema_factory
        )

        annotations = {}
        params = {"__annotations__": annotations}
        # noinspection PyDataclass
        for f in fields(annotation):
            type_ = f.type
            if f.default is not MISSING:
                type_ = Optional[type_]
                if f.default is None or isinstance(
                    f.default, (str, int, bool, float, Decimal)
                ):
                    params[f.name] = f.default
            elif f.default_factory is not MISSING:
                type_ = Optional[type_]
                params[f.name] = dataclasses.field(default_factory=f.default_factory)
            annotations[f.name] = schema_factory.get_input(type_)

        wrap_type = dataclass(type(name, tuple(), params))
        input_ = strawberry.input(wrap_type)
        schema_factory.inputs[name] = input_
        # noinspection PyDataclass
        for field in fields(input_):
            field.type = schema_factory.get_input(field.type)
        return input_


def build_resolvable_field(
    action: Action,
    schema_factory: SchemaFactory,
    key: str,
    params: Dict[str, Any],
):
    for handler_filter in schema_factory.handler_filters:
        action, continue_filtering = handler_filter.filter(action, schema_factory)
        if not continue_filtering:
            break

    fn = action.fn
    if action.batch_invoker:
        fn = _wrap_fn_in_data_loader(
            action.fn, action.batch_invoker, action.cache_control
        )
    sig = inspect.signature(fn)
    return_type = schema_factory.get_type(sig.return_annotation)
    params["__annotations__"][key] = return_type
    params[key] = strawberry.field(resolver=fn)


def _wrap_fn_in_data_loader(
    fn: Callable, batch_invoker: BatchInvoker, cache_control: CacheControlABC
):
    ttl = cache_control.ttl if hasattr(cache_control, "ttl") else 10
    data_loader = DataLoader(
        load_fn=batch_invoker.fn,
        max_batch_size=batch_invoker.max_batch_size,
        cache_map=_UserCache(relativedelta(seconds=ttl)),
    )

    sig = inspect.signature(fn)

    async def wrapper(self) -> sig.return_annotation:
        args = batch_invoker.arg_extractor(self)
        result = await data_loader.load(*args)
        return result

    return wrapper


@dataclass
class _UserCacheEntry:
    value: Any
    expire_at: datetime


class _UserCache(AbstractCache):
    def __init__(self, expire_in: relativedelta):
        self.cache: Dict[Any, _UserCacheEntry] = {}
        self.expire_in = expire_in

    def get(self, key: Any) -> Union[Any, None]:
        entry = self.cache.get(key)
        if entry and entry.expire_at > datetime.now():
            return entry.value

    def set(self, key: Any, value: Any) -> None:
        self.cache[key] = _UserCacheEntry(
            value, datetime.now() + self.expire_in
        )  # store data in the cache

    def delete(self, key: Any) -> None:
        del self.cache[key]  # delete key from the cache

    def clear(self) -> None:
        self.cache.clear()  # clear the cache



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional

import jwt
from jwt import InvalidTokenError
from marshy.types import ExternalItemType
from schemey.util import filter_none

from servey.security.authorization import Authorization, AuthorizationError
from servey.security.authorizer.authorizer_abc import AuthorizerABC


@dataclass
class JwtAuthorizer(AuthorizerABC):
    """
    Encoder that uses pyjwt and a locally stored key to work with tokens. In the AWS environment,
    it is probably better to use kms or cognito for this (Though cognito is hampered by not having
    the ability to specify custom per user scopes in the access token)
    """

    private_key: Any
    algorithm: str = "HS256"
    kid: Optional[str] = None
    iss: Optional[str] = None
    aud: Optional[str] = None

    def encode(self, authorization: Authorization) -> str:
        headers = filter_none({"kid": self.kid})
        encoded = jwt.encode(
            headers=headers,
            payload=filter_none(
                {
                    "iss": self.iss,
                    "sub": authorization.subject_id,
                    "aud": self.aud,
                    "exp": int(authorization.expire_at.timestamp())
                    if authorization.expire_at
                    else None,
                    "nbf": int(authorization.not_before.timestamp())
                    if authorization.not_before
                    else None,
                    "iat": int(datetime.now().timestamp()),
                    "scope": " ".join(authorization.scopes),
                }
            ),
            key=self.private_key,
            algorithm=self.algorithm,
        )
        return encoded

    @staticmethod
    def get_key_id(token: str) -> str:
        """May need the key id from the token header in order to figure out which key to use"""
        headers = jwt.get_unverified_header(token)
        return headers["kid"]

    def authorize(self, token: str) -> Authorization:
        try:
            decoded = jwt.decode(
                jwt=token,
                key=self.private_key,
                algorithms=["HS256", "RS256"],
                audience=self.aud,
                issuer=self.iss,
            )
            authorization = Authorization(
                subject_id=decoded.get("sub"),
                not_before=date_from_jwt(decoded, "nbf"),
                expire_at=date_from_jwt(decoded, "exp"),
                scopes=frozenset(decoded.get("scope").split(" ")),
            )
            return authorization
        except InvalidTokenError as e:
            raise AuthorizationError(e) from e


def date_from_jwt(decoded: ExternalItemType, key: str) -> Optional[datetime]:
    value = decoded.get(key)
    if value:
        return datetime.fromtimestamp(value)



from abc import abstractmethod, ABC
from typing import Iterator, TypeVar

from marshy.factory.impl_marshaller_factory import get_impls

from servey.subscription.subscription import Subscription

T = TypeVar("T")


class SubscriptionFinderABC(ABC):
    priority: int = 100

    @abstractmethod
    def find_subscriptions(self) -> Iterator[Subscription]:
        """Find all available subscriptions"""


def find_subscriptions() -> Iterator[Subscription]:
    subscription_finders = get_impls(SubscriptionFinderABC)
    for subscription_finder in subscription_finders:
        finder = subscription_finder()
        yield from finder.find_subscriptions()



"""
Standard test mechanism for actions
"""
import inspect
from unittest import TestCase

from marshy import dump

from servey.action.action import Action
from servey.action.example import Example
from servey.action.util import get_marshaller_for_params
from servey.finder.action_finder_abc import find_actions


def _define_action_example_test(action_: Action, example: Example):
    def test(self):
        marshaller = get_marshaller_for_params(action_.fn, set())
        kwargs = marshaller.load(example.params)
        result = action_.fn(**kwargs)
        sig = inspect.signature(action_.fn)
        externalized_result = dump(result, sig.return_annotation)
        self.assertEqual(example.result, externalized_result)

    return test


def define_test_class():
    test_methods = {}
    for action in find_actions():
        for example in action.examples or []:
            if example.include_in_tests:
                test_methods[
                    "test_" + action.name + "__" + example.name
                ] = _define_action_example_test(action, example)
    test_class = type("TestServeyActions", (TestCase,), test_methods)
    return test_class



from abc import abstractmethod, ABC
from typing import Dict

from celery import Celery


class CeleryConfigABC(ABC):
    @abstractmethod
    def configure(self, app: Celery, global_ns: Dict):
        """
        Configure celery
        """
